import logging.handlers as handlers, socket, os, time


class WebsockifySysLogHandler(handlers.SysLogHandler):
    """
    A handler class that sends proper Syslog-formatted messages,
    as defined by RFC 5424.
    """

    _legacy_head_fmt = '<{pri}>{ident}[{pid}]: '
    _rfc5424_head_fmt = '<{pri}>1 {timestamp} {hostname} {ident} {pid} - - '
    _head_fmt = _rfc5424_head_fmt
    _legacy = False
    _timestamp_fmt = '%Y-%m-%dT%H:%M:%SZ'
    _max_hostname = 255
    _max_ident = 24 #safer for old daemons
    _send_length = False
    _tail = '\n'


    ident = None


    def __init__(self, address=('localhost', handlers.SYSLOG_UDP_PORT),
                 facility=handlers.SysLogHandler.LOG_USER,
                 socktype=None, ident=None, legacy=False):
        """
        Initialize a handler.

        If address is specified as a string, a UNIX socket is used. To log to a
        local syslogd, "WebsockifySysLogHandler(address="/dev/log")" can be
        used. If facility is not specified, LOG_USER is used. If socktype is
        specified as socket.SOCK_DGRAM or socket.SOCK_STREAM, that specific
        socket type will be used. For Unix sockets, you can also specify a
        socktype of None, in which case socket.SOCK_DGRAM will be used, falling
        back to socket.SOCK_STREAM. If ident is specified, this string will be
        used as the application name in all messages sent. Set legacy to True
        to use the old version of the protocol.
        """

        self.ident = ident

        if legacy:
            self._legacy = True
            self._head_fmt = self._legacy_head_fmt

        handlers.SysLogHandler.__init__(self, address, facility, socktype)


    def emit(self, record):
        """
        Emit a record.

        The record is formatted, and then sent to the syslog server. If
        exception information is present, it is NOT sent to the server.
        """

        try:
            # Gather info.
            text = self.format(record).replace(self._tail, ' ')
            if not text: # nothing to log
                return

            pri = self.encodePriority(self.facility,
                                      self.mapPriority(record.levelname))

            timestamp = time.strftime(self._timestamp_fmt, time.gmtime());

            hostname = socket.gethostname()[:self._max_hostname]

            if self.ident:
                ident = self.ident[:self._max_ident]
            else:
                ident = ''

            pid = os.getpid() # shouldn't need truncation

            # Format the header.
            head = {
                'pri': pri,
                'timestamp': timestamp,
                'hostname': hostname,
                'ident': ident,
                'pid': pid,
            }
            msg = self._head_fmt.format(**head).encode('ascii', 'ignore')

            # Encode text as plain ASCII if possible, else use UTF-8 with BOM.
            try:
                msg += text.encode('ascii')
            except UnicodeEncodeError:
                msg += text.encode('utf-8-sig')

            # Add length or tail character, if necessary.
            if self.socktype != socket.SOCK_DGRAM:
                if self._send_length:
                    msg = ('%d ' % len(msg)).encode('ascii') + msg
                else:
                    msg += self._tail.encode('ascii')

            # Send the message.
            if self.unixsocket:
                try:
                    self.socket.send(msg)
                except socket.error:
                    self._connect_unixsocket(self.address)
                    self.socket.send(msg)

            else:
                if self.socktype == socket.SOCK_DGRAM:
                    self.socket.sendto(msg, self.address)
                else:
                    self.socket.sendall(msg)

        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)



from __future__ import print_function
import os
import sys

class BasePlugin(object):
    def __init__(self, src):
        self.source = src

    def lookup(self, token):
        return None


class ReadOnlyTokenFile(BasePlugin):
    # source is a token file with lines like
    #   token: host:port
    # or a directory of such files
    def __init__(self, *args, **kwargs):
        super(ReadOnlyTokenFile, self).__init__(*args, **kwargs)
        self._targets = None

    def _load_targets(self):
        if os.path.isdir(self.source):
            cfg_files = [os.path.join(self.source, f) for
                         f in os.listdir(self.source)]
        else:
            cfg_files = [self.source]

        self._targets = {}
        index = 1
        for f in cfg_files:
            for line in [l.strip() for l in open(f).readlines()]:
                if line and not line.startswith('#'):
                    try:
                        tok, target = line.split(': ')
                        self._targets[tok] = target.strip().rsplit(':', 1)
                    except ValueError:
                        print >>sys.stderr, "Syntax error in %s on line %d" % (self.source, index)
                index += 1

    def lookup(self, token):
        if self._targets is None:
            self._load_targets()

        if token in self._targets:
            return self._targets[token]
        else:
            return None


# the above one is probably more efficient, but this one is
# more backwards compatible (although in most cases
# ReadOnlyTokenFile should suffice)
class TokenFile(ReadOnlyTokenFile):
    # source is a token file with lines like
    #   token: host:port
    # or a directory of such files
    def lookup(self, token):
        self._load_targets()

        return super(TokenFile, self).lookup(token)


class BaseTokenAPI(BasePlugin):
    # source is a url with a '%s' in it where the token
    # should go

    # we import things on demand so that other plugins
    # in this file can be used w/o unnecessary dependencies

    def process_result(self, resp):
        host, port = resp.text.split(':')
        port = port.encode('ascii','ignore')
        return [ host, port ]

    def lookup(self, token):
        import requests

        resp = requests.get(self.source % token)

        if resp.ok:
            return self.process_result(resp)
        else:
            return None


class JSONTokenApi(BaseTokenAPI):
    # source is a url with a '%s' in it where the token
    # should go

    def process_result(self, resp):
        resp_json = resp.json()
        return (resp_json['host'], resp_json['port'])


class JWTTokenApi(BasePlugin):
    # source is a JWT-token, with hostname and port included
    # Both JWS as JWE tokens are accepted. With regards to JWE tokens, the key is re-used for both validation and decryption.

    def lookup(self, token):
        try:
            from jwcrypto import jwt
            import json

            key = jwt.JWK()
            
            try:
                with open(self.source, 'rb') as key_file:
                    key_data = key_file.read()
            except Exception as e:
                print("Error loading key file: %s" % str(e), file=sys.stderr)
                return None

            try:
                key.import_from_pem(key_data)
            except:
                try:
                    key.import_key(k=key_data.decode('utf-8'),kty='oct')
                except:
                    print('Failed to correctly parse key data!', file=sys.stderr)
                    return None

            try:
                token = jwt.JWT(key=key, jwt=token)
                parsed_header = json.loads(token.header)

                if 'enc' in parsed_header:
                    # Token is encrypted, so we need to decrypt by passing the claims to a new instance
                    token = jwt.JWT(key=key, jwt=token.claims)

                parsed = json.loads(token.claims)

                return (parsed['host'], parsed['port'])
            except Exception as e:
                print("Failed to parse token: %s" % str(e), file=sys.stderr)
                return None
        except ImportError as e:
            print("package jwcrypto not found, are you sure you've installed it correctly?", file=sys.stderr)
            return None

class TokenRedis(object):
    def __init__(self, src):
        self._server, self._port = src.split(":")

    def lookup(self, token):
        try:
            import redis
            import simplejson
        except ImportError as e:
            print("package redis or simplejson not found, are you sure you've installed them correctly?", file=sys.stderr)
            return None

        client = redis.Redis(host=self._server,port=self._port)
        stuff = client.get(token)
        if stuff is None:
            return None
        else:
            combo = simplejson.loads(stuff.decode("utf-8"))
            pair = combo["host"]
            return pair.split(':')



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright(c)2013 NTT corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Unit tests for websocket """
import unittest
from websockify import websocket

class FakeSocket:
    def __init__(self):
        self.data = b''

    def send(self, buf):
        self.data += buf
        return len(buf)

class AcceptTestCase(unittest.TestCase):
    def test_success(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        ws.accept(sock, {'upgrade': 'websocket',
                         'Sec-WebSocket-Version': '13',
                         'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertEqual(sock.data[:13], b'HTTP/1.1 101 ')
        self.assertTrue(b'\r\nUpgrade: websocket\r\n' in sock.data)
        self.assertTrue(b'\r\nConnection: Upgrade\r\n' in sock.data)
        self.assertTrue(b'\r\nSec-WebSocket-Accept: pczpYSQsvE1vBpTQYjFQPcuoj6M=\r\n' in sock.data)

    def test_bad_version(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Version': '5',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Version': '20',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})

    def test_bad_upgrade(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        self.assertRaises(Exception, ws.accept,
                          sock, {'Sec-WebSocket-Version': '13',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket2',
                                 'Sec-WebSocket-Version': '13',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})

    def test_missing_key(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Version': '13'})

    def test_protocol(self):
        class ProtoSocket(websocket.WebSocket):
            def select_subprotocol(self, protocol):
                return 'gazonk'

        ws = ProtoSocket()
        sock = FakeSocket()
        ws.accept(sock, {'upgrade': 'websocket',
                         'Sec-WebSocket-Version': '13',
                         'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q==',
                         'Sec-WebSocket-Protocol': 'foobar gazonk'})
        self.assertEqual(sock.data[:13], b'HTTP/1.1 101 ')
        self.assertTrue(b'\r\nSec-WebSocket-Protocol: gazonk\r\n' in sock.data)

    def test_no_protocol(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        ws.accept(sock, {'upgrade': 'websocket',
                         'Sec-WebSocket-Version': '13',
                         'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertEqual(sock.data[:13], b'HTTP/1.1 101 ')
        self.assertFalse(b'\r\nSec-WebSocket-Protocol:' in sock.data)

    def test_missing_protocol(self):
        ws = websocket.WebSocket()
        sock = FakeSocket()
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Version': '13',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q==',
                                 'Sec-WebSocket-Protocol': 'foobar gazonk'})

    def test_protocol(self):
        class ProtoSocket(websocket.WebSocket):
            def select_subprotocol(self, protocol):
                return 'oddball'

        ws = ProtoSocket()
        sock = FakeSocket()
        self.assertRaises(Exception, ws.accept,
                          sock, {'upgrade': 'websocket',
                                 'Sec-WebSocket-Version': '13',
                                 'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q==',
                                 'Sec-WebSocket-Protocol': 'foobar gazonk'})

class PingPongTest(unittest.TestCase):
    def setUp(self):
        self.ws = websocket.WebSocket()
        self.sock = FakeSocket()
        self.ws.accept(self.sock, {'upgrade': 'websocket',
                                   'Sec-WebSocket-Version': '13',
                                   'Sec-WebSocket-Key': 'DKURYVK9cRFul1vOZVA56Q=='})
        self.assertEqual(self.sock.data[:13], b'HTTP/1.1 101 ')
        self.sock.data = b''

    def test_ping(self):
        self.ws.ping()
        self.assertEqual(self.sock.data, b'\x89\x00')

    def test_pong(self):
        self.ws.pong()
        self.assertEqual(self.sock.data, b'\x8a\x00')

    def test_ping_data(self):
        self.ws.ping(b'foo')
        self.assertEqual(self.sock.data, b'\x89\x03foo')

    def test_pong_data(self):
        self.ws.pong(b'foo')
        self.assertEqual(self.sock.data, b'\x8a\x03foo')

class HyBiEncodeDecodeTestCase(unittest.TestCase):
    def test_decode_hybi_text(self):
        buf = b'\x81\x85\x37\xfa\x21\x3d\x7f\x9f\x4d\x51\x58'
        ws = websocket.WebSocket()
        res = ws._decode_hybi(buf)

        self.assertEqual(res['fin'], 1)
        self.assertEqual(res['opcode'], 0x1)
        self.assertEqual(res['masked'], True)
        self.assertEqual(res['length'], len(buf))
        self.assertEqual(res['payload'], b'Hello')

    def test_decode_hybi_binary(self):
        buf = b'\x82\x04\x01\x02\x03\x04'
        ws = websocket.WebSocket()
        res = ws._decode_hybi(buf)

        self.assertEqual(res['fin'], 1)
        self.assertEqual(res['opcode'], 0x2)
        self.assertEqual(res['length'], len(buf))
        self.assertEqual(res['payload'], b'\x01\x02\x03\x04')

    def test_decode_hybi_extended_16bit_binary(self):
        data = (b'\x01\x02\x03\x04' * 65)  # len > 126 -- len == 260
        buf = b'\x82\x7e\x01\x04' + data
        ws = websocket.WebSocket()
        res = ws._decode_hybi(buf)

        self.assertEqual(res['fin'], 1)
        self.assertEqual(res['opcode'], 0x2)
        self.assertEqual(res['length'], len(buf))
        self.assertEqual(res['payload'], data)

    def test_decode_hybi_extended_64bit_binary(self):
        data = (b'\x01\x02\x03\x04' * 65)  # len > 126 -- len == 260
        buf = b'\x82\x7f\x00\x00\x00\x00\x00\x00\x01\x04' + data
        ws = websocket.WebSocket()
        res = ws._decode_hybi(buf)

        self.assertEqual(res['fin'], 1)
        self.assertEqual(res['opcode'], 0x2)
        self.assertEqual(res['length'], len(buf))
        self.assertEqual(res['payload'], data)

    def test_decode_hybi_multi(self):
        buf1 = b'\x01\x03\x48\x65\x6c'
        buf2 = b'\x80\x02\x6c\x6f'

        ws = websocket.WebSocket()

        res1 = ws._decode_hybi(buf1)
        self.assertEqual(res1['fin'], 0)
        self.assertEqual(res1['opcode'], 0x1)
        self.assertEqual(res1['length'], len(buf1))
        self.assertEqual(res1['payload'], b'Hel')

        res2 = ws._decode_hybi(buf2)
        self.assertEqual(res2['fin'], 1)
        self.assertEqual(res2['opcode'], 0x0)
        self.assertEqual(res2['length'], len(buf2))
        self.assertEqual(res2['payload'], b'lo')

    def test_encode_hybi_basic(self):
        ws = websocket.WebSocket()
        res = ws._encode_hybi(0x1, b'Hello')
        expected = b'\x81\x05\x48\x65\x6c\x6c\x6f'

        self.assertEqual(res, expected)



from setuptools import setup, find_packages

version = '0.9.0'
name = 'websockify'
long_description = open("README.md").read() + "\n" + \
    open("CHANGES.txt").read() + "\n"

setup(name=name,
      version=version,
      description="Websockify.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4"
        ],
      keywords='noVNC websockify',
      license='LGPLv3',
      url="https://github.com/novnc/websockify",
      author="Joel Martin",
      author_email="github@martintribe.org",

      packages=['websockify'],
      include_package_data=True,
      install_requires=['numpy'],
      zip_safe=False,
      entry_points={
        'console_scripts': [
            'websockify = websockify.websocketproxy:websockify_init',
        ]
      },
    )
